/*
 * tlock.c: tlock is a program to lock the terminal with user's login password.
 * Copyright (C) 2005 2006 2007 2008 2009 2010 Prasad J Pandit.
 *
 * tlock is a free software; you can redistribute it and/or modify it under
 * the terms of GNU General Public Licence as published by Free Software
 * Foundation; either version 2 of the licence, or (at your option) any later
 * version.
 *
 * tlock is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public Licence for more
 * details.
 *
 * You should have received a copy of the GNU General Public Licence along
 * with tlock; if not, write to Free Software Foundation, Inc., 51 Franklin
 * St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <tlock.h>
#include <readpass.h>

char *prog = NULL;
static char *pass = NULL, *user = NULL;

short scount = 0;
extern int opterr, optind;

/* usage: print a short usage information */
void
usage (void)
{
    printf ("Usage: %s [OPTIONS]\n", prog);
}


/*
 * printh: display short summary of usage information and options provided by
 * tlock
 */
void
printh (void)
{
    usage ();
    printf ("\nOptions: \n");
    printf ("%-17s %s\n", "  -h --help", "display this help");
    printf ("%-17s %s\n", "  -v --version", "display version information");
    printf ("\nReport bugs to <pj.pandit@yahoo.co.in>\n");
}


/*
 * check_option: check an individual command-line option and take appropriate
 * action
 */
int
check_option (int argc, char *argv[])
{
    int n = 0, ind = 0;
    const char optstr[] = "+hv";
    struct option lopt[] = \
    {
        { "help", no_argument, NULL, 'h' },
        { "version", no_argument, NULL, 'v' },
        { 0, 0, 0, 0 }
    };

    opterr = 0;
    while ((n = getopt_long (argc, argv, optstr, lopt, &ind)) != -1)
    {
        switch (n)
        {
        case 'h':
            printh ();
            exit (0);

        case 'v':
            printf ("%s version %s\n", prog, VERSION);
            exit (0);

        default:
            usage ();
            exit (-1);
        }
    }

    return optind;
}

int
main (int argc, char *argv[])
{
    int ret = 0;

    prog = calloc ((strlen (argv[0]) + 1), sizeof (char));
    strcpy (prog, argv[0]);

    check_option (argc, argv);
    user = (char *) findname ();            /* Find user name */
    ret = lock_sys ();

    if (user)
        free (user);
    if (pass)
        free (pass);
    free (prog);

    return ret;
}

/* clrscr: clear screen */
void
clrscr (void)
{
    int n = 0;
    char *s = NULL;

    setupterm ((char *)0, STDOUT_FILENO, &n);
    s = tgetstr ("clear_screen", &s);
    tputs (s, 1, putchar);
}


/* lock_user: lock the terminal using password given by user */
int
lock_user (void)
{
    short flag = 1;
    char salt[3], saltset[65];
    char *passwd1 = NULL, *passwd2 = NULL;

    init_saltset (saltset);               /* Initialize salt set */
    generate_salt (salt, saltset);        /* generate salt to be used */
    passwd1 = readpass (PROMPT, DISPLAY, PASS_LEN);    /* read password */
    passwd2 = (char *) crypt (passwd1, salt);
    free (passwd1);
    passwd1 = calloc (strlen (passwd2) + 1, sizeof (char));
    strcpy (passwd1, passwd2);

    clrscr ();
    printf ("\n\n\t\tthis terminal has been locked by: %s\n", user);
    while (flag)
    {
        pass = readpass (PROMPT, DISPLAY, PASS_LEN);
        passwd2 = (char *) crypt (pass, salt);
        flag = checkpasswd (passwd1, passwd2);
        if (flag)
        {
            fprintf (stderr, "access denied, invalid password\n");
            free (pass);
        }
    }

    free (passwd1);
    return 0;
}


/*
 * lock_sys: lock the terminal using systems login password. This function
 * uses Linux-PAM interface to do the authentication
 */
int
lock_sys (void)
{
    pam_handle_t *pamh = NULL;
    const char *terminal = NULL;
    const char service[] = "login";
    const char ptsname[] = "console";
    struct pam_conv conv = { t_conv, NULL };

    signal (SIGINT, SIG_IGN);
    signal (SIGQUIT, SIG_IGN);
    signal (SIGTSTP, SIG_IGN);
    if (pam_start (service, user, &conv, &pamh) != PAM_SUCCESS)
        err (1, "pam initialisation failed");

    if (!getuid ())
    {   /* patch by: Milos Jakubicek <xjakub@fi.muni.cz> */
        terminal = ttyname (0);
        if (!strncmp (terminal, "/dev/pts/", 9))
            terminal = ptsname;
        if (pam_set_item (pamh, PAM_TTY, terminal) != PAM_SUCCESS)
            err (2, "pam tty initialisation failed");
    }

    clrscr ();
    printf ("\n\n\t\tthis terminal has been locked by: %s\n", user);
    while (pam_authenticate (pamh, 0) != PAM_SUCCESS)
        fprintf (stderr, "access denied, invalid password\n");

    pam_end (pamh, PAM_SUCCESS);
    return 0;
}


/*
 * t_conv: a conversation function used by PAM to communicate with the
 * application and fetch required inputs from application such as password,
 * etc
 */
int
t_conv (int msg __attribute__ ((unused)), const struct pam_message **pmsg,
        struct pam_response **rsp, void *app_data __attribute__ ((unused)))
{
    *rsp = calloc (1, sizeof (struct pam_response));
    if (*rsp == NULL)
        return (PAM_CONV_ERR);

    (*rsp)->resp_retcode = 0;
    switch ((*pmsg)->msg_style)
    {
    case PAM_PROMPT_ECHO_ON:
        (*rsp)->resp = user;
        break;

    case PAM_PROMPT_ECHO_OFF:
        (*rsp)->resp = readpass(PROMPT, DISPLAY, PASS_LEN);
    }

    return PAM_SUCCESS;
}


/* checkpasswd: compares two password string to see if they match */
int
checkpasswd (const char *pass1, const char *pass2)
{
    if (strlen (pass1) != strlen (pass2))
        return -1;

    while (*pass1 != '\0')
    {
        if (*pass1 != *pass2)
            return -1;
        pass1++;
        pass2++;
    }

    return 0;
}


/*
 * findname: return the user name of the user invoking tlock program; don't
 * forget to free memory allocated to `uname'
 */
char *
findname (void)
{
    char *uname = NULL;
    struct passwd *pw = NULL;

    pw = getpwuid (getuid ());
    uname = calloc ((strlen (pw->pw_name) + 1), sizeof (char));
    strcpy (uname, pw->pw_name);

    return uname;
}


/*
 * generate_salt: generates the salt to be used by crypt to encrypt the
 * password string
 */
void
generate_salt (char *psalt, char *saltset)
{
    time_t tt;
    char a1 = 0, a2 = 0;
    struct tm *tim = NULL;

    tt = time (&tt);
    tim = (struct tm *) localtime (&tt);
    if (tim != NULL)
        srand (tim->tm_sec);
    a1 = saltset[(rand () % scount)];
    a2 = saltset[(rand () % scount)];
    sprintf (psalt, "%c%c%c", a1, a2, '\0');
}


/*
 * init_saltset: initialise the `saltset' array with set of all possible
 * characters that can be used to generate salt
 */
void
init_saltset (char *saltset)
{
    char ch = 'A';

    /* initialising the salt character pool */
    while (ch <= 'Z')
        saltset[scount++] = ch++;
    ch = 'a';
    while (ch <= 'z')
        saltset[scount++] = ch++;
    ch = '0';
    while (ch <= '9')
        saltset[scount++] = ch++;

    saltset[scount++] = '.';
    saltset[scount++] = '/';
    saltset[scount] = '\0';
}
